import logging
import sys

logger = logging.getLogger('jbag')
logger.setLevel(logging.INFO)
log_format = logging.Formatter('%(asctime)s - %(filename)s[line:%(lineno)d] - %(levelname)s: %(message)s')
console_handler = logging.StreamHandler(stream=sys.stdout)
console_handler.setFormatter(log_format)
logger.addHandler(console_handler)



import numpy as np


def windowing(input_data, center, window_width, invert=False):
    min_ = (2 * center - window_width) / 2 + 0.5
    max_ = (2 * center + window_width) / 2 + 0.5
    factor = 255 / (max_ - min_)

    data = (input_data - min_) * factor

    data = np.where(data < 0, 0, data)
    data = np.where(data > 255, 255, data)
    data = data.astype(np.uint8)

    if invert:
        data = 255 - data

    return data



import json
import os.path
from base64 import b64encode, b64decode
from collections import OrderedDict
from typing import Union, LiteralString

import numpy as np
import pandas as pd
from numpy.lib.format import dtype_to_descr, descr_to_dtype
from openpyxl import load_workbook

from jbag import logger


def read_mat(input_file, key='scene'):
    from scipy.io import loadmat
    assert os.path.isfile(input_file), f'{input_file} does not exist or is not a file!'
    data = loadmat(input_file)[key]
    return data


def save_mat(output_file, data, key='scene'):
    from scipy.io import savemat
    ensure_output_file_dir_existence(output_file)
    savemat(output_file, {key: data})


def read_txt2list(input_file):
    assert os.path.isfile(input_file), f'{input_file} does not exist or is not a file!'

    with open(input_file, 'r') as input_file:
        return [each.strip('\n') for each in input_file.readlines()]


def write_list2txt(output_file, data_lst):
    ensure_output_file_dir_existence(output_file)
    with open(output_file, 'w') as file:
        for i in range(len(data_lst)):
            file.write(str(data_lst[i]))
            if i != len(data_lst) - 1:
                file.write('\n')


def save_nifti(output_file,
               data,
               voxel_spacing=None,
               orientation='LPI'):
    """
    Save image with nii format.

    Args:
        output_file (str or LiteralString):
        data (numpy.ndarray):
        voxel_spacing (sequence or None, optional, default=None): `tuple(x, y, z)`. Voxel spacing of each axis. If None,
            make `voxel_spacing` as `(1.0, 1.0, 1.0)`.
        orientation (str, optional, default="LPI"): "LPI" | "ARI". LPI: Left-Posterior-Inferior;
            ARI: Anterior-Right-Inferior.

    Returns:

    """
    if voxel_spacing is None:
        voxel_spacing = (1.0, 1.0, 1.0)  # replace this with your desired voxel spacing in millimeters

    match orientation:
        case 'LPI':
            affine_matrix = np.diag(list(voxel_spacing) + [1.0])
        case 'ARI':
            # calculate the affine matrix based on the desired voxel spacing and ARI orientation
            affine_matrix = np.array([
                [0, -voxel_spacing[0], 0, 0],
                [-voxel_spacing[1], 0, 0, 0],
                [0, 0, voxel_spacing[2], 0],
                [0, 0, 0, 1]
            ])
        case _:
            raise ValueError(f'Unsupported orientation {orientation}.')

    # create a NIfTI image object
    import nibabel as nib
    ensure_output_file_dir_existence(output_file)
    nii_img = nib.Nifti1Image(data, affine=affine_matrix)
    nib.save(nii_img, output_file)


def read_dicom_series(input_dir: Union[str, LiteralString]):
    from pydicom import dcmread

    assert os.path.exists(input_dir), f'{input_dir} does not exist!'

    instances = []
    for each in os.listdir(input_dir):
        if each.endswith('.dcm'):
            instances.append(each)

    instances.sort()
    images = []
    for slice_file_name in instances:
        slice_file = os.path.join(input_dir, slice_file_name)
        dicom_data = dcmread(slice_file)
        if 'PixelData' in dicom_data:
            pixel_data = dicom_data.pixel_array
            images.append(pixel_data)

    return np.stack(images)


# JSON
def np_object_hook(dct):
    """
    Convert JSON list or scalar to numpy.

    Args:
        dct (mapping):

    Returns:

    """
    if '__ndarray__' in dct:
        shape = dct['shape']
        dtype = descr_to_dtype(dct['dtype'])
        if shape:
            order = 'C' if dct['Corder'] else 'F'
            if dct['base64']:
                np_obj = np.frombuffer(b64decode(dct['__ndarray__']), dtype=dtype)
                np_obj = np_obj.copy(order=order)
            else:
                np_obj = np.asarray(dct['__ndarray__'], dtype=dtype, order=order)
            return np_obj.reshape(shape)

        if dct['base64']:
            np_obj = np.frombuffer(b64decode(dct['__ndarray__']), dtype=dtype)[0]
        else:
            t = getattr(np, dtype.name)
            np_obj = t(dct['__ndarray__'])
        return np_obj

    return dct


def read_json(input_json_file):
    """
    Read and convert `input_json_file`.

    Args:
        input_json_file (str or LiteralString):

    Returns:

    """
    assert os.path.isfile(input_json_file), f'{input_json_file} does not exist or is not a file!'

    with open(input_json_file, 'r') as json_file:
        dct = json.load(json_file, object_hook=np_object_hook)
    return dct


class NumpyJSONEncoder(json.JSONEncoder):
    def __init__(self, primitive=False, base64=True, **kwargs):
        """
        JSON encoder for `numpy.ndarray`.

        Args:
            primitive (bool, optional, default=False): Use primitive type if `True`. In primitive schema,
                `numpy.ndarray` is stored as JSON list and `np.generic` is stored as a number.
            base64 (bool, optional, default=True): Use base64 to encode.
            **kwargs:
        """
        self.primitive = primitive
        self.base64 = base64
        super().__init__(**kwargs)

    def default(self, obj):
        if isinstance(obj, (np.ndarray, np.generic)):
            if self.primitive:
                return obj.tolist()
            else:
                if self.base64:
                    data_json = b64encode(obj.data if obj.flags.c_contiguous else obj.tobytes()).decode("ascii")
                else:
                    data_json = obj.tolist()
                dct = OrderedDict(__ndarray__=data_json,
                                  dtype=dtype_to_descr(obj.dtype),
                                  shape=obj.shape,
                                  Corder=obj.flags['C_CONTIGUOUS'],
                                  base64=self.base64)
                return dct
        return super().default(obj)


def save_json(output_file, obj, primitive=False, base64=True):
    """
    Convert obj to JSON object and save as file.

    Args:
        output_file (str or LiteralString):
        obj (mapping):
        primitive (bool, optional, default=False): Use primitive type if `True`. In primitive schema, `numpy.ndarray` is
            stored as JSON list and `np.generic` is stored as a number.
        base64 (bool, optional, default=True): Use base64 to encode.

    Returns:

    """
    ensure_output_file_dir_existence(output_file)
    with open(output_file, 'w') as file:
        json.dump(obj, file, cls=NumpyJSONEncoder, **{'primitive': primitive, 'base64': base64})


def scp(dst_user, dst_host, dst_path, local_path, dst_port=None, recursive=False, send=False, receive=False):
    """
    Transmit file(s) through scp.

    Args:
        dst_user (str):
        dst_host (str):
        dst_path (str or LiteralString):
        local_path (str or LiteralString):
        dst_port (str or int or None, optional, default=None): If None, usually refer to port 22.
        recursive (bool, default=False): Transmit directories recursively.
        send (bool, default=False): Send file(s) from local to destination.
        receive (bool, default=False): Receive file(s) sent from destination.

    Returns:

    """
    assert send ^ receive

    cmd = 'scp'
    dst = f'{dst_user}@{dst_host}:{dst_path}'
    if recursive:
        cmd += ' -r'
    if dst_port:
        cmd += f' -P {dst_port}'
    if send:
        cmd += f' {local_path} {dst}'
    else:
        cmd += f' {dst} {local_path}'
    os.system(cmd)


def save_excel(output_file, data: Union[dict, pd.DataFrame], sheet_name: str = 'Sheet1', append: bool = False,
               overlay_sheet: bool = False,
               column_width: int = None, auto_adjust_width: bool = False, index=False):
    """
    Save data to Excel file.
    Args:
        output_file (str or LiteralString):
        data (dict | pd.DataFrame):
        sheet_name (str, optional, default='Sheet1'):
        append (bool, optional, default=False): If True, append to existing file.
        overlay_sheet (bool, optional, default=False): If True, overwrite existing sheet. Note that this option only works for appending mode.
        column_width (int, optional, default=None): Set column width to the given value, if exist.
        auto_adjust_width (bool, optional, default=False): If True, adjust column width according to the content length.
        index (bool, optional, default=False): If True, set index column on the work sheet.

    Returns:

    """
    write_mode = 'a' if append else 'w'

    if write_mode == 'a' and not os.path.exists(output_file):
        logger.warning(f'Try to append data to a non-existing file: {output_file}, change mode to write instead.')
        write_mode = 'w'

    if write_mode == 'a' and overlay_sheet:
        if_sheet_exists = 'overlay'
    else:
        if_sheet_exists = None

    if isinstance(data, dict):
        data = pd.DataFrame(data)

    ensure_output_file_dir_existence(output_file)

    with pd.ExcelWriter(output_file, engine='openpyxl', mode=write_mode, if_sheet_exists=if_sheet_exists) as writer:
        data.to_excel(writer, sheet_name=sheet_name, index=index)

    if column_width is not None or auto_adjust_width:
        book = load_workbook(output_file)
        worksheet = book[sheet_name]

        for column_cells in worksheet.columns:
            if column_width is not None:
                width = column_width
            else:
                max_length = max(len(str(cell.value)) for cell in column_cells)
                width = max_length + 5

            column_letter = column_cells[0].column_letter
            worksheet.column_dimensions[column_letter].width = width

        # Save the workbook
        book.save(output_file)


def ensure_output_dir_existence(output_dir):
    mk_output_dir = not os.path.exists(output_dir)
    if mk_output_dir:
        os.makedirs(output_dir)
    return mk_output_dir, output_dir


def ensure_output_file_dir_existence(output_file):
    output_dir = os.path.split(output_file)[0]
    return ensure_output_dir_existence(output_dir)



from docx import Document
from docx.enum.text import WD_PARAGRAPH_ALIGNMENT


def add_paragraph(doc: Document,
                  text=None,
                  alignment: WD_PARAGRAPH_ALIGNMENT = WD_PARAGRAPH_ALIGNMENT.JUSTIFY,
                  style=None
                  ):
    if text is None:
        text = ''
    p = doc.add_paragraph(text=text, style=style)
    if alignment is not None:
        p.alignment = alignment
    return p



import os

from jbag.io import ensure_output_file_dir_existence


def dicom2nifti(input_dicom_series, output_nifti_file, pydicom_read_force=False):
    import dicom2nifti as d2n

    assert os.path.exists(input_dicom_series), f'{input_dicom_series} does not exist!'

    if pydicom_read_force:
        d2n.settings.pydicom_read_force = pydicom_read_force

    ensure_output_file_dir_existence(output_nifti_file)
    d2n.convert_directory(input_dicom_series, output_nifti_file)



import torch

from jbag.transforms.transforms import RandomTransform


class MirrorTransform(RandomTransform):
    def __init__(self, keys,
                 apply_probability,
                 allowed_axes,
                 p_per_axes: float = 0.5):
        """
        Mirror transform.
        Args:
            keys (str or sequence):
            apply_probability (float):
            allowed_axes (int or sequence): Axis(es) for mirroring.
            p_per_axes (bool, optional, default=False): Probability for performing transform on each axis.

        """
        super().__init__(keys, apply_probability)
        if isinstance(allowed_axes, int):
            allowed_axes = [allowed_axes]
        self.allowed_axes = allowed_axes
        self.p_per_axes = p_per_axes

    def _call_fun(self, data):
        allowed_axes = [axis for axis in self.allowed_axes if torch.rand(1) < self.p_per_axes]
        if len(allowed_axes) == 0:
            return data
        for key in self.keys:
            value: torch.Tensor = data[key]
            value = torch.flip(value, allowed_axes)
            data[key] = value
        return data



from docx.shared import Pt
from docx.text.paragraph import Paragraph


def add_text(paragraph: Paragraph, text: str, font: str = 'Times New Roman', size: int = 10, bold=False, italic=False,
             underline=False, subscript=False, superscript=False):
    run = paragraph.add_run()
    run.font.name = font
    run.font.size = Pt(size)
    run.font.bold = bold
    run.font.italic = italic
    run.font.underline = underline
    run.font.subscript = subscript
    run.font.superscript = superscript
    run.text = text



from typing import Union

from docx.oxml import OxmlElement
from docx.oxml.ns import qn
from docx.shared import Pt
from docx.table import _Cell


def set_cell(cell: _Cell, text, font='Times New Roman', font_size=10, bold=False, italic=False,
             underline=False):
    cell.text = text
    run = cell.paragraphs[0].runs[0]
    run.font.name = font
    run.font.size = Pt(font_size)
    run.font.bold = bold
    run.font.italic = italic
    run.font.underline = underline


def set_cell_border(cell: _Cell,
                    borders: Union[str, list[str], tuple[float, ...]],
                    styles: Union[str, list[str], tuple[float, ...]] = 'single',
                    sizes: Union[float, list[float], tuple[float, ...]] = 4,
                    colors: Union[str, list[str], tuple[str, ...]] = 'auto'):
    """

    Args:
        cell:
        borders: Option: top, bottom, left, right
        styles:
        sizes: 4 is equal to 0.5pt in word software

    Returns:

    """
    if isinstance(borders, str):
        borders = [borders]

    valid_borders = ['top', 'bottom', 'left', 'right']
    for border in borders:
        assert border in valid_borders, f'Invalid border: {border}'

    if isinstance(styles, str):
        styles = [styles] * len(borders)
    if not isinstance(sizes, (list, tuple)):
        sizes = [sizes] * len(borders)
    if isinstance(colors, str):
        colors = [colors] * len(borders)

    assert len(borders) == len(styles) == len(sizes) == len(colors)

    tc = cell._tc
    tcPr = tc.get_or_add_tcPr()
    tcBorders = OxmlElement('w:tcBorders')

    for border, style, size, color in zip(borders, styles, sizes, colors):
        border_element = OxmlElement(f'w:{border}')
        border_element.set(qn('w:val'), style)
        border_element.set(qn('w:sz'), str(size))
        border_element.set(qn('w:color'), color)
        tcBorders.append(border_element)
    tcPr.append(tcBorders)



from typing import Union

import torch

from jbag.transforms._utils import get_non_one_scalar
from jbag.transforms.transforms import RandomTransform


class MultiplicativeBrightnessTransform(RandomTransform):
    def __init__(self, keys,
                 apply_probability,
                 multiplier_range: Union[tuple[float], list[float]],
                 synchronize_channels: bool = False,
                 p_per_channel: float = 1):
        """
        Brightness transform.
        Args:
            keys (str or sequence):
            apply_probability (float):
            multiplier_range (sequence): Multiplier for brightness adjustment is sampled from this range without value of `1` if `1` is in range.
            synchronize_channels (bool, optional, default=False): If True, use the same parameters for all channels.
            p_per_channel (float, optional, default=1): Probability of applying transform to each channel.
        """
        assert len(multiplier_range) == 2 and multiplier_range[1] >= multiplier_range[0]
        super().__init__(keys, apply_probability)
        self.multiplier_range = multiplier_range
        self.synchronize_channels = synchronize_channels
        self.p_per_channel = p_per_channel

    def _call_fun(self, data):
        apply_to_channel = torch.where(torch.rand(len(self.keys)) < self.p_per_channel)[0]
        if len(apply_to_channel) == 0:
            return data
        if self.synchronize_channels:
            multipliers = [get_non_one_scalar(self.multiplier_range), ] * len(apply_to_channel)
        else:
            multipliers = [get_non_one_scalar(self.multiplier_range) for _ in range(len(apply_to_channel))]

        for c, m in zip(apply_to_channel, multipliers):
            value = data[self.keys[c]]
            value *= m
            data[self.keys[c]] = value
        return data



import numpy as np
import torch.nn as nn


class DeepSupervisionLossWrapper(nn.Module):
    def __init__(self, loss_criterion, weights=None):
        """
        Loss criterion wrapper for deep supervision.
        Args:
            loss_criterion (torch.nn.Module): Loss criterion function.
            weights (sequence[float, ...], optional, default=None): Weights for losses of different scale outputs.
        """
        super().__init__()
        if weights is not None:
            assert any([x != 0 for x in weights])
        self.loss_criterion = loss_criterion
        self.weights = weights

    def forward(self, *args):
        assert all([isinstance(arg, (list, tuple)) for arg in args])

        if self.weights is None:
            weights = [1] * len(args[0])
        else:
            weights = self.weights

        return sum(
            [weights[i] * self.loss_criterion(*inputs) for i, inputs in enumerate(zip(*args)) if weights[i] != 0])


def get_deep_supervision_loss_weights(network_config):
    deep_supervision_scales = get_deep_supervision_scales(network_config)
    weights = np.array([1 / (2 ** i) for i in range(len(deep_supervision_scales))])
    return weights


def get_deep_supervision_scales(network_config):
    if 'deep_supervision' not in network_config or not network_config['deep_supervision']:
        return None
    strides = network_config['strides']
    deep_supervision_scales = list(list(i) for i in 1 / np.cumprod(np.vstack(strides), axis=0))[:-1]
    return deep_supervision_scales


def set_deep_supervision(network, flag):
    if hasattr(network, 'decoder'):
        if hasattr(network.decoder, 'deep_supervision'):
            network.decoder.deep_supervision = flag



from typing import Type

import torch.nn as nn
from torch.nn.modules.conv import _ConvNd


def get_conv_op(dim):
    match dim:
        case 1:
            return nn.Conv1d
        case 2:
            return nn.Conv2d
        case 3:
            return nn.Conv3d
        case _:
            raise ValueError()


def get_norm_op(op_name, dim):
    match op_name:
        case 'InstanceNorm':
            match dim:
                case 2:
                    return nn.InstanceNorm2d
        case 'BatchNorm':
            match dim:
                case 2:
                    return nn.BatchNorm2d


def get_non_linear_op(op_name):
    match op_name:
        case 'leaky_relu':
            return nn.LeakyReLU
        case 'relu':
            return nn.ReLU


def get_matching_conv_transpose_op(conv_op: Type[_ConvNd]):
    match conv_op:
        case nn.Conv1d:
            return nn.ConvTranspose1d
        case nn.Conv2d:
            return nn.ConvTranspose2d
        case nn.Conv3d:
            return nn.ConvTranspose3d
        case _:
            raise ValueError(f'Unknown conv op {conv_op}')


def get_conv_dimensions(conv_op: Type[_ConvNd]):
    match conv_op:
        case nn.Conv1d:
            return 1
        case nn.Conv2d:
            return 2
        case nn.Conv3d:
            return 3
        case _:
            raise ValueError(f'Unknown conv op {conv_op}')



from abc import ABC, abstractmethod

import numpy as np

from .utils import get_margin, encode_one_hot


class CoordinateGenerator(ABC):

    @abstractmethod
    def __getitem__(self, index):
        ...

    @abstractmethod
    def __len__(self):
        ...

    @abstractmethod
    def regenerate(self):
        ...

    @staticmethod
    def get_margin_mask(image_shape: tuple | list, patch_size: tuple | list):
        """
        Build a mask that cover the area which can be view as the central point of the patch with the size of
        patch_size. Mask the central area that can be extracted. Left the margin area.

        Args:
            image_shape (sequence):
            patch_size (sequence):

        Returns:

        """
        assert len(image_shape) == len(patch_size)
        margin = get_margin(patch_size)
        mask = np.zeros(image_shape, dtype=bool)
        mask[tuple([slice(j[0], i - j[1]) for i, j in zip(image_shape, margin)])] = 1
        return mask


class WeightedCoordinateGenerator(CoordinateGenerator):

    def __init__(self, num_coordinates, patch_size, weight_map):
        """
        Choose coordinates according to the weight.

        Args:
            num_coordinates (int):
            patch_size (sequence):
            weight_map (numpy.ndarray or torch.Tensor):
        """
        # number of coordinates that need to be picked.
        self.n = num_coordinates
        self.patch_size = patch_size
        self.weight_map = weight_map
        self.coordinates = self.get_coordinates()

    def __getitem__(self, index):
        return self.coordinates[index]

    def __len__(self):
        return self.n

    def get_coordinates(self):
        weight_map = self.weight_map / np.sum(self.weight_map).astype(np.float64)
        weight_map = weight_map.flatten()
        coordinates = np.random.choice(weight_map.size, size=self.n, p=weight_map)
        coordinates = np.asarray(np.unravel_index(coordinates, self.weight_map.shape))
        return coordinates.T

    def regenerate(self):
        self.coordinates = self.get_coordinates()


class BalancedCoordinateGenerator(WeightedCoordinateGenerator):
    def __init__(self, num_coordinates, data, patch_size):
        """
        Generate coordinates according to the proportion of categories. In general, the smaller the percentage of
        category, the more likely it is to be taken. Parameter data is the un-one-hot label.

        Args:
            num_coordinates (int):
            data (numpy.ndarray or torch.Tensor): The label matrix. this data will be needed to generate the weight map.
            patch_size (sequence):
        """
        self.n = num_coordinates
        self.patch_size = patch_size
        super().__init__(num_coordinates=num_coordinates, patch_size=patch_size, weight_map=self.get_weight_map(data))

    def get_weight_map(self, data):
        # one-hot first
        weight_map = encode_one_hot(data)

        # exclude margin
        margin_mask = self.get_margin_mask(data.shape, self.patch_size)
        weight_map = weight_map * margin_mask
        weight_map = weight_map.astype(np.float32)

        # normalize on every category axis
        category_sum = np.sum(weight_map, axis=tuple(range(1, len(data.shape) + 1)), keepdims=True)
        weight_map = weight_map / (category_sum + 1)
        weight_map = np.sum(weight_map, axis=0).astype(np.float32)
        return weight_map


class GridCoordinateGenerator(CoordinateGenerator):
    def __init__(self, original_shape, patch_shape, valid_shape):
        """

        Args:
            original_shape (sequence): The shape of original image.
            patch_shape (sequence): The shape of one patch.
            valid_shape (sequence): The size of valid patch. Valid shape should be less than patch shape. A valid patch
                needs to be extracted from a patch.
        """
        self.original_shape = np.asarray(original_shape, dtype=np.uint32)
        self.patch_shape = np.asarray(patch_shape, dtype=np.uint32)
        self.valid_shape = np.asarray(valid_shape, dtype=np.uint32)

        self.padding_size = get_margin(patch_shape) - get_margin(valid_shape)
        self.valid_central_coordinates = self.get_coordinates()
        self.padded_coordinates = self.valid_central_coordinates + self.padding_size[:, 0][None, :]

    def get_coordinates(self):
        """
        Get central coordinates according to the original shape and valid shape.

        Returns:

        """
        coordinate_dims = [range(0, i, j) for i, j in zip(self.original_shape, self.valid_shape)]
        coordinates = np.meshgrid(*coordinate_dims, indexing='ij')
        coordinates = list(map(lambda x: x.flatten(), coordinates))
        # get the corner coordinate of each patch
        coordinates = np.stack(coordinates, axis=-1)
        coordinates = np.where(coordinates + self.valid_shape > self.original_shape,
                               self.original_shape - self.valid_shape, coordinates)

        # shift coordinate to patch center
        shift = get_margin(self.valid_shape)[:, 0]
        central_coordinates = coordinates + shift
        return central_coordinates.astype(np.int32)

    def __getitem__(self, index):
        return self.padded_coordinates[index]

    def __len__(self):
        return len(self.padded_coordinates)

    def regenerate(self):
        ...



from typing import Union

import SimpleITK as sitk


def overlay(image, label_map, colors: Union[list, tuple] = None, opacity=0.5):
    image = sitk.GetImageFromArray(image)
    label_map = sitk.GetImageFromArray(label_map)
    if colors:
        if len(colors) == 1:
            itk_color_map = list(colors[0])
        else:
            itk_color_map = list(colors[-1])
            for i in colors[:-1]:
                itk_color_map += list(i)
        overlaid_image = sitk.LabelOverlay(image, label_map, opacity=opacity, colormap=itk_color_map)
    else:
        overlaid_image = sitk.LabelOverlay(image, label_map, opacity=opacity)
    overlaid_image = sitk.GetArrayFromImage(overlaid_image)
    return overlaid_image



import datetime
from enum import Enum
from random import randint

from pydicom import Dataset
from pydicom.dataset import FileDataset, FileMetaDataset
from pydicom.uid import generate_uid


class Modality(Enum):
    CT = 1


def get_dicom_dataset(name, modality: Modality) -> FileDataset:
    file_meta = FileMetaDataset()
    file_meta.TransferSyntaxUID = '1.2.840.10008.1.2.1'
    file_meta.ImplementationVersionName = 'DICOM'

    ds = FileDataset(name, {}, file_meta=file_meta, preamble=b'\0' * 128)
    ds.is_implicit_VR = False
    ds.is_little_endian = True
    ds.ImageType = ['DERIVED', 'SECONDARY']

    # add properties
    patient(ds)
    general_study(ds)
    patient_study(ds)
    frame_of_reference(ds)
    general_equipment(ds)
    general_image(ds)
    general_acquisition(ds)
    image_plane(ds)
    image_pixel(ds)
    SOP_common(ds)
    VOI_LUT(ds)
    general_series(ds, file_meta, modality)

    # set datatime
    dt = datetime.datetime.now()
    date_str = dt.strftime('%Y%m%d')
    time_str = dt.strftime('%H%M%S.%f')  # long format with micro seconds

    ds.ContentDate = date_str
    ds.ContentTime = time_str
    ds.StudyDate = date_str
    ds.StudyTime = time_str
    ds.SeriesDate = date_str
    ds.SeriesTime = time_str
    ds.AcquisitionDate = date_str
    ds.AcquisitionTime = time_str
    ds.InstanceCreationDate = date_str
    ds.InstanceCreationTime = time_str

    ds.RescaleIntercept = ''
    ds.RescaleSlope = ''

    return ds


def patient(ds: Dataset):
    ds.PatientName = 'Patient Name'
    ds.PatientID = '1'
    ds.PatientSex = ''
    ds.PatientBirthDate = ''


def general_study(ds: Dataset):
    ds.StudyInstanceUID = generate_uid()
    ds.StudyDescription = ''
    ds.ReferringPhysicianName = ''


def patient_study(ds: Dataset):
    ds.PatientAge = ''
    ds.PatientWeight = ''


def general_series(ds: Dataset, file_meta: FileMetaDataset, modality: Modality):
    ds.Modality = modality.name
    match modality:
        case Modality.CT:
            file_meta.MediaStorageSOPClassUID = '1.2.840.10008.5.1.4.1.1.2'
            ds.SOPClassUID = '1.2.840.10008.5.1.4.1.1.2'
            CT_image(ds)

    ds.SeriesInstanceUID = generate_uid(None)
    ds.SeriesNumber = str(randint(1000, 9999))
    ds.ProtocolName = 'DICOM'
    ds.PatientPosition = ''
    ds.AccessionNumber = '123456'


def frame_of_reference(ds: Dataset):
    ds.FrameOfReferenceUID = generate_uid(None)


def general_equipment(ds: Dataset):
    ds.Manufacturer = ''
    ds.InstitutionName = 'INSTITUTION_NAME_UNDEFINED'
    ds.ManufacturerModelName = ''
    ds.SoftwareVersions = ''


def general_image(ds: Dataset):
    ds.InstanceNumber = ''
    ds.PatientOrientation = ''
    ds.ContentDate = ''
    ds.ContentTime = ''
    ds.ImageType = ['SECONDARY', 'DERIVED']
    ds.LossyImageCompression = '00'


def general_acquisition(ds: Dataset):
    ds.AcquisitionNumber = ''
    ds.AcquisitionDate = ''
    ds.AcquisitionTime = ''


def image_plane(ds: Dataset):
    ds.PixelSpacing = ''
    ds.ImageOrientationPatient = ['1', '0', '0', '0', '1', '0']
    ds.ImagePositionPatient = ['0', '0', '0']
    ds.SliceThickness = ''
    ds.SpacingBetweenSlices = ''
    ds.SliceLocation = ''


def image_pixel(ds: Dataset):
    ds.Rows = 0
    ds.Columns = 0

    ds.BitsAllocated = 0
    ds.BitsStored = 0
    ds.HighBit = 0

    ds.PixelRepresentation = 0

    ds.SmallestImagePixelValue = ''
    ds.LargestImagePixelValue = ''

    ds.PixelData = ''


def SOP_common(ds: Dataset):
    ds.SOPClassUID = ''
    ds.SOPInstanceUID = ''

    ds.SpecificCharacterSet = 'ISO_IR 100'
    ds.InstanceCreationDate = ''
    ds.InstanceCreationTime = ''

    ds.InstanceCreatorUID = ''


def VOI_LUT(ds: Dataset):
    ds.WindowCenter = ''
    ds.WindowWidth = ''


def CT_image(ds: Dataset):
    ds.SamplesPerPixel = 1

    ds.PhotometricInterpretation = 'MONOCHROME2'

    ds.BitsAllocated = 16
    ds.BitsStored = 12
    ds.HighBit = ds.BitsStored - 1



import multiprocessing as mp
from time import sleep

from tqdm import tqdm


def fork(fn, n_workers, params):
    """
    Invoke fn with multiple processors.
    Args:
        fn:
        n_workers:
        params:

    Returns:

    """
    r = []
    with mp.get_context('fork').Pool(n_workers) as p:
        for each_param in params:
            if not isinstance(each_param, tuple):
                each_param = tuple(each_param)
            r.append(p.starmap_async(fn, (each_param,)))

        remaining = list(range(len(params)))
        workers = [e for e in p._pool]

        with tqdm(desc=None, total=len(params)) as pbar:
            while len(remaining) > 0:
                all_alive = all([e.is_alive() for e in workers])
                if not all_alive:
                    raise RuntimeError(
                        'One of your background processes is missing. In that case reducing the number of workers might help.')

                done = [i for i in remaining if r[i].ready()]
                for _ in done:
                    pbar.update()

                remaining = [i for i in remaining if i not in done]
                sleep(0.1)

    results = [i.get()[0] for i in r]
    return results



import os.path

from docx import Document


def import_style_from_template(doc: Document, template_doc_file):
    if not os.path.exists(template_doc_file):
        raise FileNotFoundError(template_doc_file)

    template_doc = Document(template_doc_file)

    for style in template_doc.styles:
        doc.styles.add_style(style.style_id, style.type)



from typing import Union

import numpy as np
import torch
import torch.nn.functional as F

from jbag.samplers.coordinate_generator import GridCoordinateGenerator
from jbag.samplers.utils import central_crop, get_margin


class GridSampler:
    def __init__(self, data: Union[np.ndarray, torch.Tensor], patch_size, valid_size=None):
        """
        Perform grid sample and recovery. The last N dimensions of given data will be split.

        Args:
            data (numpy.ndarray or torch.Tensor): The data needs to be sample.
            patch_size (sequence): The data will be divided into patches with the shape of patch_shape.
            valid_size (sequence or None, optional, default=None): Valid patch size, used for restore. If `None`, as
                same as `patch_size`.
        """
        self.patch_size = np.asarray(patch_size, dtype=np.uint32)
        self.valid_size = np.asarray(valid_size, dtype=np.uint32) if valid_size is not None else self.patch_size
        assert len(self.patch_size.shape) == len(self.valid_size.shape)

        self.full_size = data.shape
        # shape of data, but batch and/or channel dimensions are excluded
        self.original_shape = self.full_size[-len(self.patch_size):]
        self.coordinate_generator = GridCoordinateGenerator(self.original_shape, patch_shape=self.patch_size,
                                                            valid_shape=self.valid_size)

        if not (self.patch_size == self.valid_size).all():
            if isinstance(data, torch.Tensor):
                pad_size = tuple(np.flip(self.coordinate_generator.padding_size, axis=0).flatten())
                self.padded_data = F.pad(data, pad=pad_size, mode='constant', value=0)
            else:
                pad_size = self.coordinate_generator.padding_size
                if len(self.full_size) != len(self.valid_size):
                    extra_dimensions = [[0, 0]] * (len(self.full_size) - len(self.valid_size))
                    pad_size = np.concatenate((np.asarray(extra_dimensions, dtype=np.int32), pad_size))
                self.padded_data = np.pad(data, pad_width=pad_size, mode='constant', constant_values=0)
        else:
            self.padded_data = data
        self.index = 0

    def restore(self, patches, restore_shape=None):
        """
        Every element's batch/channel dimensions in blocks should be put at the beginning.

        Args:
            patches (sequence):
            restore_shape (sequence or None, optional, default=None): Restore array shape. Batch/channel dimensions
                of restored array can differ from the data which was divided as long as keeping the same image
                dimensions. If None, as same as `self.full_size`.
        """
        patches = list(map(lambda x: self._shrink_shape(x), patches))
        if not restore_shape:
            restore_shape = self.full_size
        return self._rebuild_image(patches, restore_shape)

    def _shrink_shape(self, data: Union[np.ndarray, torch.Tensor]):
        """
        Shrink data to assigned shape.

        Args:
            data (numpy.ndarray or torch.Tensor):

        Returns:

        """
        if (self.patch_size == self.valid_size).all():
            return data
        center = [e // 2 if e % 2 != 0 else e // 2 - 1 for e in self.valid_size]
        center = np.asarray(center)
        center = center + self.coordinate_generator.padding_size[:, 0]
        cropped = central_crop(data, center, self.valid_size)
        return cropped

    def _rebuild_image(self, patches, restore_shape):
        assert len(patches) > 0
        assert len(self.coordinate_generator.valid_central_coordinates) == len(patches)
        result = np.zeros(shape=restore_shape, dtype=patches[0].dtype) if isinstance(patches[0], np.ndarray) \
            else torch.zeros(size=restore_shape, dtype=patches[0].dtype, device=patches[0].device)

        block_margin = get_margin(self.valid_size)
        margin = block_margin[:, 0][None, :]
        coordinates = self.coordinate_generator.valid_central_coordinates - margin
        extra_dimensions = [slice(None)] * (len(restore_shape) - len(self.valid_size))
        for i, block in enumerate(patches):
            slice_sequence = extra_dimensions + \
                             [slice(coordinates[i, j], coordinates[i, j] + self.valid_size[j])
                              for j in range(len(self.valid_size))]

            result[tuple(slice_sequence)] = block
        return result

    def __len__(self):
        return len(self.coordinate_generator)

    def __iter__(self):
        return self

    def __next__(self):
        if self.index >= len(self):
            raise StopIteration

        patch = central_crop(self.padded_data, self.coordinate_generator[self.index], self.patch_size)
        self.index += 1
        return patch



import setuptools

setuptools.setup(
    name='jbag',
    version='4.0.7',
    author='Dai Jian',
    author_email='daijian@stumail.ysu.edu.cn',
    description='Tools for medical image processing and deap learning.',
    long_description_content_type='text/markdown',
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3.11',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
    ],
    python_requires='~=3.11',
    install_requires=[
        'numpy >= 1.21.0',
        'nibabel >= 4.0.1',
        'scipy >= 1.7.3',
        'SimpleITK >= 2.1.1.2',
        'pydicom >= 2.4.3',
        'dicom2nifti >= 2.4.8',
        'pandas',
        'openpyxl',
        'python-docx',
        'scikit-image',
    ]
)



import numpy as np
import torch

from jbag.transforms.transforms import Transform


class ToType(Transform):
    def __init__(self, keys, dtype):
        super().__init__(keys)
        self.dtype = dtype

    def _call_fun(self, data):
        for key in self.keys:
            value = data[key].astype(self.dtype)
            data[key] = value
        return data


class ToTensor(Transform):
    def __init__(self, keys):
        super().__init__(keys)

    def _call_fun(self, data):
        for key in self.keys:
            value = data[key]
            if isinstance(value, np.ndarray):
                data[key] = torch.from_numpy(value)
            else:
                data[key] = torch.tensor(value)
        return data


class AddChannel(Transform):
    def __init__(self, keys, axis):
        """
        Add additional dimension in specific position.

        Args:
            keys (str or sequence):
            axis (int):
        """
        super().__init__(keys)
        self.axis = axis

    def _call_fun(self, data):
        for key in self.keys:
            value = data[key]
            value = np.expand_dims(value, axis=self.axis)
            data[key] = value
        return data


class Repeat(Transform):
    def __init__(self, keys, repeats, axis):
        """
        Repeat specific dimension.

        Args:
            keys (str or sequence):
            axis (int):
            repeats (int):
        """
        super().__init__(keys)
        self.axis = axis
        self.repeats = repeats

    def _call_fun(self, data):
        for key in self.keys:
            value = data[key]
            value = np.repeat(value, self.repeats, axis=self.axis)
            data[key] = value
        return data



from typing import Union, Type

import torch
import torch.nn as nn
from torch.nn.modules.conv import _ConvNd
from torch.nn.modules.dropout import _DropoutNd

from jbag.config import Config
from jbag.models.network_weight_initialization import initialize_network
from jbag.models.utils import get_conv_op, get_norm_op, get_non_linear_op, get_matching_conv_transpose_op, \
    get_conv_dimensions


class UNet(nn.Module):
    def __init__(self, input_channels: int,
                 num_classes: int,
                 num_stages: int,
                 num_features_per_stage: Union[int, list[int], tuple[int, ...]],
                 conv_op: Type[_ConvNd],
                 kernel_sizes: Union[int, list[int], tuple[int, ...]],
                 strides: Union[int, list[int], tuple[int, ...]],
                 num_conv_per_stage_encoder: Union[int, list[int], tuple[int, ...]],
                 num_conv_per_stage_decoder: Union[int, list[int], tuple[int, ...]],
                 conv_bias: bool = False,
                 norm_op: Union[None, Type[nn.Module]] = None,
                 norm_op_kwargs: dict = None,
                 dropout_op: Union[None, Type[_DropoutNd]] = None,
                 dropout_op_kwargs: dict = None,
                 non_linear: Union[None, Type[nn.Module]] = None,
                 non_linear_kwargs: dict = None,
                 deep_supervision: bool = False,
                 non_linear_first: bool = False
                 ):
        super().__init__()

        if isinstance(num_conv_per_stage_encoder, int):
            num_conv_per_stage_encoder = [num_conv_per_stage_encoder] * num_stages

        if isinstance(num_conv_per_stage_decoder, int):
            num_conv_per_stage_decoder = [num_conv_per_stage_decoder] * (num_stages - 1)

        assert len(num_conv_per_stage_encoder) == num_stages
        assert len(num_conv_per_stage_decoder) == num_stages - 1
        self.encoder = Encoder(input_channels=input_channels,
                               num_stages=num_stages,
                               num_features_per_stage=num_features_per_stage,
                               conv_op=conv_op,
                               kernel_sizes=kernel_sizes,
                               strides=strides,
                               num_conv_per_stage=num_conv_per_stage_encoder,
                               conv_bias=conv_bias,
                               norm_op=norm_op,
                               norm_op_kwargs=norm_op_kwargs,
                               dropout_op=dropout_op,
                               dropout_op_kwargs=dropout_op_kwargs,
                               non_linear=non_linear,
                               non_linear_kwargs=non_linear_kwargs,
                               return_skips=True,
                               non_linear_first=non_linear_first
                               )
        self.decoder = Decoder(num_classes=num_classes,
                               encoder=self.encoder,
                               num_conv_per_stage=num_conv_per_stage_decoder,
                               deep_supervision=deep_supervision,
                               non_linear_first=non_linear_first
                               )

    def forward(self, x):
        skips = self.encoder(x)
        return self.decoder(skips)


class Encoder(nn.Module):
    def __init__(self, input_channels: int,
                 num_stages: int,
                 num_features_per_stage: Union[int, list[int], tuple[int, ...]],
                 conv_op: Type[_ConvNd],
                 kernel_sizes: Union[int, list[int], tuple[int, ...]],
                 strides: Union[int, list[int], tuple[int, ...]],
                 num_conv_per_stage: Union[int, list[int], tuple[int, ...]],
                 conv_bias: bool = False,
                 norm_op: Union[None, Type[nn.Module]] = None,
                 norm_op_kwargs: dict = None,
                 dropout_op: Union[None, Type[_DropoutNd]] = None,
                 dropout_op_kwargs: dict = None,
                 non_linear: Union[None, Type[nn.Module]] = None,
                 non_linear_kwargs: dict = None,
                 return_skips: bool = False,
                 non_linear_first: bool = False
                 ):
        super().__init__()
        if isinstance(kernel_sizes, int):
            kernel_sizes = [kernel_sizes] * num_stages
        if isinstance(num_features_per_stage, int):
            num_features_per_stage = [num_features_per_stage] * num_stages
        if isinstance(num_conv_per_stage, int):
            num_conv_per_stage = [num_conv_per_stage] * num_stages
        if isinstance(strides, int):
            strides = [strides] * num_stages

        assert len(kernel_sizes) == num_stages
        assert len(num_features_per_stage) == num_stages
        assert len(num_conv_per_stage) == num_stages
        assert len(strides) == num_stages

        stages = []
        for i in range(num_stages):
            conv_stride = strides[i]
            stages.append(StackedConvBlock(input_channels=input_channels,
                                           output_channels=num_features_per_stage[i],
                                           num_convs=num_conv_per_stage[i],
                                           conv_op=conv_op,
                                           kernel_size=kernel_sizes[i],
                                           initial_stride=conv_stride,
                                           conv_bias=conv_bias,
                                           norm_op=norm_op,
                                           norm_op_kwargs=norm_op_kwargs,
                                           dropout_op=dropout_op,
                                           dropout_op_kwargs=dropout_op_kwargs,
                                           non_linear=non_linear,
                                           non_linear_kwargs=non_linear_kwargs,
                                           non_linear_first=non_linear_first
                                           ))
            input_channels = num_features_per_stage[i]

        self.stages = nn.ModuleList(stages)
        self.return_skips = return_skips
        self.output_channels = num_features_per_stage
        self.conv_op = conv_op
        self.kernel_sizes = kernel_sizes
        self.strides = strides
        self.conv_bias = conv_bias
        self.norm_op = norm_op
        self.norm_op_kwargs = norm_op_kwargs
        self.dropout_op = dropout_op
        self.dropout_op_kwargs = dropout_op_kwargs
        self.non_linear = non_linear
        self.non_linear_kwargs = non_linear_kwargs

    def forward(self, x):
        ret = []
        for layer in self.stages:
            x = layer(x)
            ret.append(x)
        if self.return_skips:
            return ret
        else:
            return ret[-1]


class Decoder(nn.Module):
    def __init__(self, num_classes: int,
                 encoder: Encoder,
                 num_conv_per_stage: Union[int, list[int], tuple[int, ...]],
                 deep_supervision: bool,
                 norm_op: Union[None, Type[nn.Module]] = None,
                 norm_op_kwargs: dict = None,
                 dropout_op: Union[None, Type[_DropoutNd]] = None,
                 dropout_op_kwargs: dict = None,
                 non_linear: Union[None, Type[nn.Module]] = None,
                 non_linear_kwargs: dict = None,
                 conv_bias: bool = None,
                 non_linear_first: bool = False
                 ):
        super().__init__()
        self.deep_supervision = deep_supervision
        num_stages_encoder = len(encoder.output_channels)
        if isinstance(num_conv_per_stage, int):
            num_conv_per_stage = [num_conv_per_stage] * (num_stages_encoder - 1)
        assert len(num_conv_per_stage) == num_stages_encoder - 1
        conv_transpose_op = get_matching_conv_transpose_op(encoder.conv_op)
        conv_bias = encoder.conv_bias if conv_bias is None else conv_bias
        norm_op = encoder.norm_op if norm_op is None else norm_op
        norm_op_kwargs = encoder.norm_op_kwargs if norm_op_kwargs is None else norm_op_kwargs
        dropout_op = encoder.dropout_op if dropout_op is None else dropout_op
        dropout_op_kwargs = encoder.dropout_op_kwargs if dropout_op_kwargs is None else dropout_op_kwargs
        non_linear = encoder.non_linear if non_linear is None else non_linear
        non_linear_kwargs = encoder.non_linear_kwargs if non_linear_kwargs is None else non_linear_kwargs

        stages = []
        conv_transpose_ops = []
        seg_layers = []
        for i in range(1, num_stages_encoder):
            input_features_below = encoder.output_channels[-i]
            input_features_skip = encoder.output_channels[-(i + 1)]
            stride_for_transpose_conv = encoder.strides[-i]
            conv_transpose_ops.append(conv_transpose_op(input_features_below,
                                                        input_features_skip,
                                                        stride_for_transpose_conv,
                                                        stride_for_transpose_conv,
                                                        bias=conv_bias
                                                        ))

            stages.append(StackedConvBlock(
                input_channels=2 * input_features_skip,
                output_channels=input_features_skip,
                num_convs=num_conv_per_stage[i - 1],
                conv_op=encoder.conv_op,
                kernel_size=encoder.kernel_sizes[-(i + 1)],
                initial_stride=1,
                conv_bias=conv_bias,
                norm_op=norm_op,
                norm_op_kwargs=norm_op_kwargs,
                dropout_op=dropout_op,
                dropout_op_kwargs=dropout_op_kwargs,
                non_linear=non_linear,
                non_linear_kwargs=non_linear_kwargs,
                non_linear_first=non_linear_first
            ))

            seg_layers.append(
                encoder.conv_op(input_features_skip, num_classes, kernel_size=1, stride=1, padding=0, bias=True))

        self.stages = nn.ModuleList(stages)
        self.conv_transpose_ops = nn.ModuleList(conv_transpose_ops)

        # Deep supervision is added anywhere, even for deep_supervision = False.
        # This allows load pre-trained parameters with and without deep supervision simultaneously.
        self.seg_layers = nn.ModuleList(seg_layers)

    def forward(self, skips):
        low_resolution_features = skips[-1]
        seg_outputs = []
        for i in range(len(self.stages)):
            x = self.conv_transpose_ops[i](low_resolution_features)
            x = torch.cat((x, skips[-(i + 2)]), dim=1)
            x = self.stages[i](x)
            if self.deep_supervision:
                seg_outputs.append(self.seg_layers[i](x))
            elif i == (len(self.stages) - 1):
                seg_outputs.append(self.seg_layers[-1](x))

            low_resolution_features = x

        seg_outputs = seg_outputs[::-1]

        if not self.deep_supervision:
            return seg_outputs[0]
        else:
            return seg_outputs


class StackedConvBlock(nn.Sequential):
    def __init__(self, input_channels: int,
                 output_channels: int,
                 num_convs: int,
                 conv_op: Type[_ConvNd],
                 kernel_size: Union[int, list[int], tuple[int, ...]],
                 initial_stride: Union[int, list[int], tuple[int, ...]],
                 conv_bias: bool = False,
                 norm_op: Union[None, Type[nn.Module]] = None,
                 norm_op_kwargs: dict = None,
                 dropout_op: Union[None, Type[_DropoutNd]] = None,
                 dropout_op_kwargs: dict = None,
                 non_linear: Union[None, Type[nn.Module]] = None,
                 non_linear_kwargs: dict = None,
                 non_linear_first: bool = False
                 ):
        if not isinstance(output_channels, (list, tuple)):
            output_channels = [output_channels] * num_convs

        super().__init__(ConvDropoutNormReLU(input_channels=input_channels,
                                             output_channels=output_channels[0],
                                             conv_op=conv_op,
                                             kernel_size=kernel_size,
                                             stride=initial_stride,
                                             conv_bias=conv_bias,
                                             norm_op=norm_op,
                                             norm_op_kwargs=norm_op_kwargs,
                                             dropout_op=dropout_op,
                                             dropout_op_kwargs=dropout_op_kwargs,
                                             non_linear=non_linear,
                                             non_linear_kwargs=non_linear_kwargs,
                                             non_linear_first=non_linear_first),
                         *[
                             ConvDropoutNormReLU(input_channels=output_channels[i - 1],
                                                 output_channels=output_channels[i],
                                                 conv_op=conv_op,
                                                 kernel_size=kernel_size,
                                                 stride=1,
                                                 conv_bias=conv_bias,
                                                 norm_op=norm_op,
                                                 norm_op_kwargs=norm_op_kwargs,
                                                 dropout_op=dropout_op,
                                                 dropout_op_kwargs=dropout_op_kwargs,
                                                 non_linear=non_linear,
                                                 non_linear_kwargs=non_linear_kwargs,
                                                 non_linear_first=non_linear_first)
                             for i in range(1, num_convs)
                         ]
                         )


class ConvDropoutNormReLU(nn.Sequential):
    def __init__(self,
                 input_channels: int,
                 output_channels: int,
                 conv_op: Type[_ConvNd],
                 kernel_size: Union[int, list[int], tuple[int, ...]],
                 stride: Union[int, list[int], tuple[int, ...]],
                 conv_bias: bool = False,
                 norm_op: Union[None, Type[nn.Module]] = None,
                 norm_op_kwargs: dict = None,
                 dropout_op: Union[None, Type[_DropoutNd]] = None,
                 dropout_op_kwargs: dict = None,
                 non_linear: Union[None, Type[nn.Module]] = None,
                 non_linear_kwargs: dict = None,
                 non_linear_first: bool = False
                 ):

        if norm_op_kwargs is None:
            norm_op_kwargs = {}
        if non_linear_kwargs is None:
            non_linear_kwargs = {}

        ops = []

        conv_dim = get_conv_dimensions(conv_op)
        if not isinstance(kernel_size, (list, tuple)):
            kernel_size = [kernel_size] * conv_dim

        conv = conv_op(
            input_channels,
            output_channels,
            kernel_size,
            stride,
            padding=[(i - 1) // 2 for i in kernel_size],
            dilation=1,
            bias=conv_bias
        )

        ops.append(conv)

        if dropout_op is not None:
            dropout = dropout_op(**dropout_op_kwargs)
            ops.append(dropout)

        if norm_op is not None:
            norm = norm_op(output_channels, **norm_op_kwargs)
            ops.append(norm)

        if non_linear is not None:
            non_linear = non_linear(**non_linear_kwargs)
            ops.append(non_linear)

        if non_linear_first and (norm_op is not None and non_linear is not None):
            ops[-1], ops[-2] = ops[-2], ops[-1]

        super().__init__(*ops)


def build_unet(network_config: Config):
    conv_op = get_conv_op(network_config.conv_dim)
    norm_op = get_norm_op(network_config.norm_op, network_config.conv_dim)
    non_linear_op = get_non_linear_op(network_config.non_linear)

    params = {'input_channels': network_config.input_channels,
              'num_classes': network_config.num_classes,
              'num_stages': network_config.num_stages,
              'num_features_per_stage': network_config.num_features_per_stage,
              'conv_op': conv_op,
              'kernel_sizes': network_config.kernel_sizes,
              'strides': network_config.strides,
              'num_conv_per_stage_encoder': network_config.num_conv_per_stage_encoder,
              'num_conv_per_stage_decoder': network_config.num_conv_per_stage_decoder,
              'conv_bias': network_config.conv_bias,
              'norm_op': norm_op,
              'norm_op_kwargs': network_config.norm_op_kwargs,
              'non_linear': non_linear_op,
              'non_linear_kwargs': network_config.non_linear_kwargs,
              'deep_supervision': network_config.deep_supervision,
              'non_linear_first': network_config.non_linear_first,
              }
    network = UNet(**params)

    initialize_network(network, network_config)

    return network



import random
from abc import ABC, abstractmethod
from itertools import cycle

from jbag.samplers.coordinate_generator import BalancedCoordinateGenerator
from jbag.samplers.patch_picker import WeightedPatchPicker


class PreloadDataset(ABC):
    def __init__(self, samples: list,
                 patch_size,
                 n_patches_per_sample,
                 n_samples_alive=1,
                 shuffle=True,
                 transforms=None):
        """
        Store the subjects on a queue of length queue_length in the RAM.

        Args:
            samples (list): List of samples.
            patch_size (sequence):
            n_patches_per_sample (int): Number of patches cropped from each sample.
            n_samples_alive (int, optional, default=1): Maximum number of samples saved in RAM, should in the range of
                [1, len(samples)].
            shuffle (bool, optional, default=True): If `True`, shuffle the samples.
            transforms (torchvision.transforms.Compose or None):
        """
        if shuffle:
            random.shuffle(samples)
        self.sample_iterator = cycle(samples)

        self.patch_size = patch_size
        self.n_patches_per_sample = n_patches_per_sample

        if n_samples_alive <= 0:
            n_samples_alive = 1
        elif n_samples_alive > len(samples):
            n_samples_alive = len(samples)
        self.n_samples_alive = n_samples_alive

        self.sample_queue = [None, ] * self.n_samples_alive
        self.index = 0

        self.transforms = transforms

    def update_queue(self, pos: int):
        """
        Update the queue element in the given position.

        Args:
            pos (int):

        Returns:

        """
        sample_idx = next(self.sample_iterator)

        if self.sample_queue[pos] and self.sample_queue[pos][0] == sample_idx:
            # only update the patch picker.
            _, patch_picker = self.sample_queue[pos]
            patch_picker.reset()
        else:
            data = self.get_sample(sample_idx)
            coordinate_generator = BalancedCoordinateGenerator(self.n_patches_per_sample,
                                                               self.get_weight_map(data),
                                                               self.patch_size)

            patch_picker = WeightedPatchPicker(data, self.patch_size, coordinate_generator)
            self.sample_queue[pos] = (sample_idx, patch_picker)

    @abstractmethod
    def get_sample(self, index):
        """
        Get sample by index from `self.sample_iterator`.

        Args:
            index (object): Object like str or int that indices the sample.

        Returns:

        """
        ...

    @abstractmethod
    def get_weight_map(self, data):
        ...

    def __iter__(self):
        return self

    def __next__(self):
        index = self.index
        self.index = (self.index + 1) % self.n_samples_alive

        # only invoke when visiting the index at the first time.
        if self.sample_queue[index] is None:
            self.update_queue(index)
        try:
            data = self.next(index)
        except StopIteration:
            self.update_queue(index)
            data = self.next(index)
        return data

    def next(self, index):
        _, patch_picker = self.sample_queue[index]
        data = next(patch_picker)
        if self.transforms:
            data = self.transforms(data)
        return data



from collections import OrderedDict
from typing import Union, Iterable

import numpy as np
import torch

from .coordinate_generator import CoordinateGenerator
from .utils import central_crop


class WeightedPatchPicker:
    def __init__(self, data: Union[tuple, list, dict, np.ndarray, torch.Tensor],
                 patch_size,
                 coordinates: CoordinateGenerator):
        """
        An iterable for picking a patch from data. The original data can be a single numpy.ndarray or a set of data.

        Args:
            data (sequence or mapping or numpy.ndarray or torch.Tensor):
            patch_size (sequence):
            coordinates (CoordinateGenerator):
        """
        self.data = data
        self.patch_size = patch_size
        self.index = 0
        self.coordinates = coordinates

    def reset(self):
        self.coordinates.regenerate()
        self.index = 0

    def __iter__(self):
        return self

    def __next__(self):
        index = self.index
        self.index += 1

        if index >= len(self):
            raise StopIteration

        if isinstance(self.data, Iterable):
            # if the data is a bundle of samples, all ndarray data will be cropped according the candidate coordinate.
            if isinstance(self.data, dict):
                results = OrderedDict()
                for k, v in self.data.items():
                    if isinstance(v, np.ndarray):
                        results[k] = central_crop(v, self.coordinates[index], self.patch_size)
                    else:
                        results[k] = v
                return results

            results = []
            for data in self.data:
                if isinstance(data, np.ndarray):
                    results.append(central_crop(data, self.coordinates[index], self.patch_size))
                else:
                    results.append(data)
            return tuple(results)
        return central_crop(self.data, self.coordinates[index], self.patch_size)

    def __len__(self):
        return len(self.coordinates)
